% Entropy Inpainting for RGB Images - 04/28/2012
% Jean-Marc Berthommé
%
% - 03/01/12:
%   . inpainting based on information theory
% - 03/08/12:
%   . basis extension tests
%   . different themes & filling methods
% - 04/25/12:
%   . color RGB inpainting final implementation reviewed
% - 04/28/12:
%   . follow the total entropy, i.e. known + unknown parts

function inpaint_rgb
scr = get(0,'ScreenSize'); % screen size
N = 3;                     % patch size: NxN decomposition -> entropy
the = 'bsc6';              % theme: bsc6/puz6/lcx6/rnd6/uk24/cha24
nrm = 'L2';                % norm for the cache: L1/L2/Linf
dwd = 1;                   % display words: 0/1
dbg = 0;                   % debug   mode : 0/1
[img,c] = data(the,nrm);   % image 'img'/ cache 'c'
[h,w,b3] = size(img);      % height 'h'  / width 'w' / RGB bit depth 'b3'
b = b3/3;                  % bit depth 'b' per channel

% plot considerations
nRow = 5;                          % total nb of lines           | 6 - 7
nCol = 9;                          % total nb of columns         | 7 - 12
nBor = 4;                          % nb of borders for the words | 3 - 6
[Xp,Yp] = meshgrid(1:nCol,1:nRow); % subplot array: nRow x nCol
Idp = nCol*(Yp-1)+Xp;              % indices array
Inp = Xp > nBor;                   % indices "inside"
mWin = Idp(Inp);                   % main window
bWin = Idp(~Inp);                  % border window (left)
t = true(h,w);                     % true cache 't'
fig3 = 0; fig4 = 0;                % figure 3 & 4 flags

%-------------------------------------------------------------------------%
[Pimg, Dimg] = prob(img,N,t); himg = Hp(Pimg); % original image 'img'
disp_img(scr,nRow,nCol,mWin,bWin,N, himg,img,Dimg,Pimg, dwd,dbg);

[Pdam, Ddam] = prob(img,N,c); hdam = Hp(Pdam); % damaged image 'dam'
mid = false(h,w,b3); dam = false(h,w,b3); ndam = size(Ddam,1);
mid(:,:,b)=1; mid(:,:,2*b)=1; mid(:,:,3*b)=1;  % ~> gray (0.5)
dam(repmat(c, [1 1 b3])) = img(repmat(c, [1 1 b3]));
dam(repmat(~c, [1 1 b3])) = mid(repmat(~c, [1 1 b3]));
disp_dam(scr,nRow,nCol,mWin,bWin,N, hdam,dam,Ddam,Pdam, dwd,dbg);

inp = dam; Dinp = Ddam; Pinp = Pdam;           % image to inpaint 'inp'
ninp = ndam; hinp = hdam;                      % initialized from 'dam'
%-------------------------------------------------------------------------%
[X,Y] = meshgrid(1:w,1:h);
se1 = strel('diamond',1); se2 = strel('square',2*(N-1)+1);

nh = sum(~c(:))+1; kh = 1;              % nb of holes & counter
Hinp = zeros(1,nh); Htot = zeros(1,nh); % inpaint and global image H tables
Hinp(kh)=hinp; Htot(kh)=hinp+b3*sum(~c(:)); kh=kh+1; % 1st registration

while ~isequal(c, t)
  % current layer 'lay' to fill in
  e = imerode(~c,se1); d = imdilate(~c & ~e, se2); lay = d & ~e;
  fig4 = disp_lay(scr, fig4, lay, dbg);
  
  % NxN squares indexes
  Id = id_dec(X, N, lay); Yl = Y(Id(:,1)); Xl = X(Id(:,1));
  nl = size(Yl,1);       % nb of words to fill in the layer
  Up = sum(~c(Id),2);    % unknown pixels in the different words
  Idl = (1:nl)'; kl = 1; % layer indexes & 1st word index in 'lay'
  
  while sum(Up) > 0
    % find the next word index 'kl'
    Dl = sq_dist([Yl(kl),Xl(kl)], [Yl(~~Up),Xl(~~Up)]);
    Dp = Dl'/max(Dl); Kp = Up(~~Up)/N^2; % distance & knowledge pct
    [~,idl] = min(Dp + Kp);              % nearest & surest neighbor
    Idl_ = Idl(~~Up); kl = Idl_(idl);    % ~> next word to select
    
    % kl-th word to inpaint in the layer 'lay'
    if Up(kl) >= 1                    % at least 1 unknown pixel
      s = sel(kl,Yl,Xl,N,t);          % selection 's'
      k = c(s)';                      % 'k' known pixels in 's'
      v = inp(repmat(c&s,[1 1 b3]))'; % 'v' known values in 's'
      nk = sum(k); nu = sum(~k);      % nb of 'k' known & 'u' unknown px
      fig3 = disp_inp(scr, fig3, nRow, nCol, mWin, bWin, N, ...
                     kh, hinp, inp, Dinp, Pinp, dwd, dbg, 0, kl, Yl, Xl);
      
      Dk = reshape(Dinp(repmat(k,ninp,b3)),ninp,b3*nk); % dictionary 'k'
      wm = mean(~xor(Dk, repmat(v,ninp,1)), 2);         % word match score
      th = sort(unique(wm),'descend');                  % thresholds set
      i  = 1; while sum(wm >= th(i)) < min(2,ninp), i=i+1; end;
      wd = wm >= th(i); % probable words map
      Dt = Dinp(wd,:);  % test dictionnary 't'
      nt = sum(wd);     % nb of candidates to test
      
      a=all(~xor(repmat(v,nt,1),reshape(Dt(repmat(k,nt,b3)),nt,b3*nk)),2);
      
      if sum(a) == 1
        word = Dt(a,:);
      elseif sum(a) == 0
        word = false(1,b3*N^2);
        word(repmat( k,1,b3)) = v;
        word(repmat(~k,1,b3)) = (wm(wd)/sum(wm(wd)))' * ... % weights
                            reshape(Dt(repmat(~k,nt,b3)), nt,b3*nu) >= 0.5;
      else
        Ht = zeros(nt,1);
        for i = 1:nt
          inp(repmat(s, [1 1 b3])) = Dt(i,:)';
          pt = prob(inp, N, c|s); Ht(i) = Hp(pt);
        end
        [~,idmin] = min(abs(Ht-hdam));
        word = Dt(idmin,:);
      end
      
      inp(repmat(s, [1 1 b3])) = word';     % inpaint
      c = c | s; Up = sum(~c(Id),2);        % update cache
      [Pinp, Dinp] = prob(inp,N,c);         % count words
      hinp = Hp(Pinp); ninp = size(Dinp,1); % calculate entropy
      
      Hinp(kh) = hinp; Htot(kh) = hinp + b3*sum(~c(:)); kh=kh+1; % register
      disp_inp(scr, fig3, nRow, nCol, mWin, bWin, N, ...        % display
               kh, hinp, inp, Dinp, Pinp, dwd, dbg, 0, kl, Yl, Xl);
    end
  end
end
disp_inp(scr, fig3, nRow, nCol, mWin, bWin, N, ... % final display
          kh-1, hinp, inp, Dinp, Pinp, dwd, dbg, 1, kl, Yl, Xl);
disp_ent(scr, kh, Hinp, Htot);
record(the, b, img, dam, inp);

% clear all; close all;
fprintf('Press any key to continue...\n');
pause; clear all; close all;

%******************************* Functions *******************************%
function [img,c] = data(theme, nrm)
% Import the data, i.e. the image 'img' and the cache 'c'
dir = 'data_in';
if strcmp(theme,'bsc6')        % -> blind spot - experiment (c) - 6-bit /
    mat = sprintf('%s/%s.mat', dir, theme);  % 2-bit per RGB channel
    load(mat);                 
    img = labels;
elseif strcmp(theme,'puz6')    % -> puzzle 6-bit / 2-bit per RGB channel
    mat = sprintf('%s/%s.mat', dir, theme);
    load(mat);
    img = labels;
    [h,w,~] = size(img);
    c = cache(h,w,nrm);
elseif strcmp(theme,'lcx6') % -> little cross 6-bit / 2-bit per RGB channel
    mat = sprintf('%s/%s.mat', dir, theme);
    load(mat);
    img = labels;
    [h,w,~] = size(img);
    c = cache(h,w,nrm);
elseif strcmp(theme,'rnd6')    % -> random 6-bit / 2-bit per RGB channel
    RandStream.setDefaultStream(...
        RandStream('mt19937ar','seed',sum(100*clock)));
    s = 40; h = s; w = s;
    img = round(rand(h,w,6));
    c = cache(h,w,nrm);
elseif strcmp(theme,'uk24')    % -> UK 24-bit / 8-bit per RGB channel
    fi = sprintf('%s/uk24.png', dir);
    im = imread(fi); [h,w,~] = size(im);
    img_r = logical(reshape(de2bi(im(:,:,1)),h,w,8));
    img_g = logical(reshape(de2bi(im(:,:,2)),h,w,8));
    img_b = logical(reshape(de2bi(im(:,:,3)),h,w,8));
    img = cat(3,img_r,img_g,img_b);
    c = cache(h,w,nrm);
elseif strcmp(theme,'cha24')   % -> Charlie 24-bit / 8-bit per RGB channel
    fi = sprintf('%s/cha24.jpg', dir);
    im = imread(fi); [h,w,~] = size(im);
    img_r = logical(reshape(de2bi(im(:,:,1)),h,w,8));
    img_g = logical(reshape(de2bi(im(:,:,2)),h,w,8));
    img_b = logical(reshape(de2bi(im(:,:,3)),h,w,8));
    img = cat(3,img_r,img_g,img_b);
    c = cache(h,w,nrm);
else                           % -> exception
    error('The theme %s is not defined', theme);
end

function c = cache(h, w, nrm)
% Build the cache 'c'
[X,Y] = meshgrid(1:w,1:h);
if     strcmp(nrm,'L1'),
    c = abs(X-(w+1)/2) + abs(Y-(h+1)/2) > 0.29*(h^2+w^2)^0.5;
elseif strcmp(nrm,'L2')
    c = (X-(w+1)/2).^2 + (Y-(h+1)/2).^2 > 0.04*(h^2+w^2); % 0.07
elseif strcmp(nrm,'Linf')
    c = max( abs(X-(w+1)/2), abs(Y-(h+1)/2)) > 0.2*(h^2+w^2)^0.5;
end

% function [img,c] = data_(theme)
% % Import the data, i.e. the image 'img' and the cache 'c', from:
% % http://yokoya.naist.jp/research/inpainting
% dir = 'input_bmp';
% fi = sprintf('%s/%s.bmp', dir, theme);
% fc = sprintf('%s/a%s.bmp', dir, theme);
% 
% I = imread(fi); C = imread(fc); [h,w,~] = size(I);
% Ir = logical(reshape(de2bi(I(:,:,1)),h,w,8));
% Ig = logical(reshape(de2bi(I(:,:,2)),h,w,8));
% Ib = logical(reshape(de2bi(I(:,:,3)),h,w,8));
% 
% img = cat(3, Ir,Ig,Ib);
% c   = ~(C(:,:,1) == 255 & C(:,:,2) == 0 & C(:,:,3) == 0);

function record(the, b, img, dam, inp)
dir = 'data_out';
s_img = sprintf('%s/%s_%s.png',dir,the,'img');
s_dam = sprintf('%s/%s_%s.png',dir,the,'dam');
s_inp = sprintf('%s/%s_%s.png',dir,the,'inp');

imwrite( uint8( map(img) * (2^8-1)/(2^b-1) ), s_img);
imwrite( uint8( map(dam) * (2^8-1)/(2^b-1) ), s_dam);
imwrite( uint8( map(inp) * (2^8-1)/(2^b-1) ), s_inp);

function s = sel(kl, Yl, Xl, N, t)
% Build the selection 's' (binary image)
s = ~t; s(Yl(kl):Yl(kl)+N-1, Xl(kl):Xl(kl)+N-1) = 1;

function [img_,b] = map(img)
% Calculate the image values on [0,2^b-1] for the 3 RGB channels
[h,w,b3] = size(img); b = b3/3;
img_r = zeros(h,w); img_g = zeros(h,w); img_b = zeros(h,w);
for i=1:b
    img_r = img_r + 2^(i-1)*img(:,:,b*(1-1)+i); % 1st channel
    img_g = img_g + 2^(i-1)*img(:,:,b*(2-1)+i); % 2nd channel
    img_b = img_b + 2^(i-1)*img(:,:,b*(3-1)+i); % 3rd channel
end
img_ = cat(3, img_r, img_g, img_b);

function disp_words(B, P, N, b, nRow, nCol, bWin, dbg)
% Display the words of an image fully decomposed with their respective %
% NB: We can't display more than size(bWin,1) words!
if dbg, Idx = (1:size(B,1))';     % normal ranking for debugging
else [~,Idx] = sort(P,'descend'); % first most probable words for release
end
for i=1:min(size(B,1), size(bWin,1))
    Ri = reshape(B(Idx(i),:),N,N,3*b);
    stit_i = sprintf('%0.1f%%', 100*P(Idx(i)));
    subplot(nRow,nCol,bWin(i)); image(map(Ri)/(2^b-1));
    axis image; axis off; title(stit_i,'Color','w');
end

function disp_img(scr, nRow, nCol, mWin, bWin, N, ...
                   himg, img, Dimg, Pimg, dwd, dbg)
% Display the original image
nw = size(Dimg,1); [img_,b] = map(img);
h1 = sprintf('%d-bit RGB Image - %d words\n', 3*b, nw);
s1 = sprintf('%sHimgels = %0.4f bits - %d x %d x %d', h1, himg, N, N, 3*b);

f1 = figure(1); set(gcf,'Color',[0.2,0.2,0.2]);
set(f1,'Position',[scr(3)/3, scr(4)/2, scr(3)/3, scr(4)/2-50*2]);
subplot(nRow,nCol,mWin); image(img_/(2^b-1));
axis image; axis off; title(s1,'Color','w');
if dwd, disp_words(Dimg, Pimg, N, b, nRow, nCol, bWin, dbg); end;

function disp_dam(scr, nRow, nCol, mWin, bWin, N, ...
                   hdam, dam, Ddam, Pdam, dwd, dbg)
% Display the damaged image
nw = size(Ddam,1); [dam_,b] = map(dam);
h2 = sprintf('Damaged Image - %d words\n', nw);
s2 = sprintf('%sHdam = %0.4f bits - %d x %d x %d', h2, hdam, N, N, 3*b);

f2 = figure(2); set(gcf,'Color',[0.2,0.2,0.2]);
set(f2,'Position',[2*scr(3)/3, scr(4)/2, scr(3)/3, scr(4)/2-50*2]);
subplot(nRow,nCol,mWin); image(dam_/(2^b-1));
axis image; axis off; title(s2,'Color','w');
if dwd, disp_words(Ddam, Pdam, N, b, nRow, nCol, bWin, dbg); end;

function fig3 = disp_inp(scr, fig3, nRow, nCol, mWin, bWin, N, ...
                      kh, hinp, inp, Dinp, Pinp, dwd, dbg, fin, kl, Yl, Xl)
% Display the inpainted image
if ~fig3, fig3 = 1; f3 = figure(3); set(gcf,'Color',[0.2,0.2,0.2]);
  set(f3,'Position',[2*scr(3)/3, 0, scr(3)/3, scr(4)/2-50*2]); end;
if ~fin && dbg
  nw = size(Dinp,1); [inp_,b] = map(inp);
  h3 = sprintf('Inpainted Image - %d words\n', nw);
  s3 = sprintf('%sk = %d - Hinp = %0.4f bits - %d x %d x %d', h3, kh, ...
                                                          hinp, N, N, 3*b);
  figure(3); subplot(nRow,nCol,mWin); image(inp_/(2^b-1));
  axis image; axis off; title(s3,'Color','w');
  if dwd, disp_words(Dinp, Pinp, N, b, nRow, nCol, bWin, dbg); end;
  disp_rect(kl,N,Yl,Xl, 'r', nRow,nCol,mWin, 3,4);
  
elseif ~fin && ~dbg && rand(1,1) > 0.95
  nw = size(Dinp,1); [inp_,b] = map(inp);
  h3 = sprintf('Inpainted Image - %d words\n', nw);
  s3 = sprintf('%sk = %d - Hinp = %0.4f bits - %d x %d x %d', h3, kh, ...
                                                          hinp, N, N, 3*b);
  figure(3); subplot(nRow,nCol,mWin); image(inp_/(2^b-1));
  axis image; axis off; title(s3,'Color','w');
  if dwd, disp_words(Dinp, Pinp, N, b, nRow, nCol, bWin, dbg); end;
  
elseif fin
  nw = size(Dinp,1); [inp_,b] = map(inp);
  h3 = sprintf('Inpainted Image - %d words\n', nw);
  s3 = sprintf('%sk = %d - Hinp = %0.4f bits - %d x %d x %d', h3, kh, ...
                                                          hinp, N, N, 3*b);
  figure(3); subplot(nRow,nCol,mWin); image(inp_/(2^b-1));
  axis image; axis off; title(s3,'Color','w');
  if dwd, disp_words(Dinp, Pinp, N, b, nRow, nCol, bWin, dbg); end;
else % do_nothing()
end

function fig4 = disp_lay(scr, fig4, lay, dbg)
% Display the current layer
if dbg
  if ~fig4, fig4 = 1; f4 = figure(4); set(gcf,'Color',[0.2,0.2,0.2]);
    set(f4,'Position',[scr(3)/3, 0, scr(3)/3, scr(4)/2-50*2]); end;
  figure(4); image(repmat(lay,[1 1 3])); axis image; axis off;
  title('Current layer','color','w');
end

function disp_ent(scr, kh, Hinp, Htot)
% Display the evolution of the entropy along the inpainting process
f5 = figure(5); set(gcf,'Color',[0.2,0.2,0.2]);
set(f5,'Position',[scr(3)/3, 0, scr(3)/3, scr(4)/2-50*2]);
nh = kh-1; c1 = [0.2 0.1 1]; c2 = [1 0.1 0]; % nb of elements & colors

subplot(2,1,1);
hold on; plot(1:nh,Hinp(1:nh),'color',c1,'linewidth',1);
plot(1:nh,Htot(1:nh),'color',c2,'linewidth',1); hold off;
axis([1 nh 0 max(0.5, 1.1*max(Hinp))]);
xlabel('Inpainting Process - Iteration k','color','w');
ylabel('H_k [bits]','color','w'); legend('H_{inp}','H_{tot}');
title('Inpainted & Total image entropy evolution','color','w');

subplot(2,1,2);
hold on; plot(1:nh,Hinp(1:nh),'color',c1,'linewidth',1);
plot(1:nh,Htot(1:nh),'color',c2,'linewidth',1); hold off;
axis([1 nh 0 1.05*max(Htot)]);
xlabel('Inpainting Process - Iteration k','color','w');
ylabel('H_k [bits]','color','w'); legend('H_{inp}','H_{tot}');

function disp_rect(kl, N, Yl, Xl, col, nRow, nCol, mWin, id1, id2)
y = [Yl(kl) Yl(kl)+N-1]; x = [Xl(kl) Xl(kl)+N-1];
rect(y, x, col, nRow, nCol, mWin, id1);
rect(y, x, col, nRow, nCol, mWin, id2);

function rect(y, x, col, nRow, nCol, mWin, win_id)
d  = 0.5;
Py = [y(1)-d;y(1)-d;y(2)+d;y(2)+d;y(1)-d];
Px = [x(1)-d;x(2)+d;x(2)+d;x(1)-d;x(1)-d];

if     win_id == 2, figure(2); subplot(nRow,nCol,mWin);
elseif win_id == 3, figure(3); subplot(nRow,nCol,mWin);
elseif win_id == 4, figure(4);
else   error('Wrong window Id %', win_id);
end
hold on; plot(Px,Py,'color',col,'linewidth', 1); hold off;

function Id = id_dec(I, N, c)
% Index decomposition of a binary image on NxNxb square patches.
% Note that here it's a sliding decomposition with redundant information
% between the square binary patches
%
% I : Boolean image (at any bit depth 'b').
% N : patch size for the image decomposition on N x N.
% c : binary cache: "1" if a pixel belongs to the cache, "0" if not.
%
% Id: indexes of the N x N x b patches to consider

[h, w, b] = size(I); % height, width, bit depth
% [X,Y] = meshgrid(1:w,1:h);
% Ind = h*(X-1)+Y; % vectorized image indexes

if N <= min(h, w)
    % construction of the NxNxb indices set row by row
    Id0 = (1:h-N+1)';
    Id1 = zeros(h-N+1, N);
    for j = 1:N
        Id1(:,j) = Id0 + j-1;
    end

    Id2 = zeros(h-N+1, N^2);
    for j = 1:N
       Id2(:,(j-1)*N+1:(j-1)*N+N) = Id1 + (j-1)*h; 
    end

    Id3 = zeros((w-N+1)*(h-N+1), N^2);
    for i = 1:w-N+1
        Id3((i-1)*(h-N+1)+1 : (i-1)*(h-N+1)+h-N+1, :) = Id2 + (i-1)*h;
    end

    Id4 = zeros((w-N+1)*(h-N+1), b*N^2);
    for k = 1:b
       Id4(:, (k-1)*N^2+1 : (k-1)*N^2 + N^2) = Id3 + (k-1)*(h*w);
    end

    % I(Id4) is the exhaustive set of all the NxNxb binary words of I
    % These words must be inside the cache (check on "c(Id3)"):
    Ok = all(c(Id3),2);
    Id = Id4(Ok,:);
else
    error('Decomposition by %dx%d is too big!', N, N);
end

function D2 = sq_dist(X,Y)
% Euclidean square distances
nx = size(X,1); ny = size(Y,1);
D2 = sum(X.^2,2)*ones(1,ny) + ones(nx,1)*sum(Y.^2,2)' - 2*(X*Y');

%******************************** Entropy ********************************%
function [P, D, L] = prob(I, N, c)
% Distribution of the NxNxb square patches inside a logical image.
% Note that here there's a sliding decomposition with redundant information
% between the patches.
%
% I : logical image (at any bit depth 'b').
% N : patch size for the image decomposition on N x N x b.
% c : binary cache: "1" if a pixel belongs to the cache, "0" if not.
%
% P : occurrence probabilities of the N x N x b patches, also called words.
% D : dictionnary of the linearized (in row) N x N x b word signatures.
% L : labels index of the decomposition of I. Eeach L(i) matches to D(i,:).

[h, w, b] = size(I); % height, width, bit depth
% [X,Y] = meshgrid(1:w,1:h);
% Ind = h*(X-1)+Y; % vectorized image indexes

if N <= min(h, w)
    % construction of the NxNxb indices set row by row
    Id0 = (1:h-N+1)';
    Id1 = zeros(h-N+1, N);
    for j = 1:N
        Id1(:,j) = Id0 + j-1;
    end

    Id2 = zeros(h-N+1, N^2);
    for j = 1:N
       Id2(:,(j-1)*N+1:(j-1)*N+N) = Id1 + (j-1)*h; 
    end

    Id3 = zeros((w-N+1)*(h-N+1), N^2);
    for i = 1:w-N+1
        Id3((i-1)*(h-N+1)+1 : (i-1)*(h-N+1)+h-N+1, :) = Id2 + (i-1)*h;
    end

    Id4 = zeros((w-N+1)*(h-N+1), b*N^2);
    for k = 1:b
       Id4(:, (k-1)*N^2+1 : (k-1)*N^2 + N^2) = Id3 + (k-1)*(h*w);
    end

    % I(Id4) is the exhaustive set of all the NxNxb binary words of I
    % These words must be inside the cache (check on "c(Id3)"):
    %  -> Id4(all(c(Id3),2),:) vs Id4
    [D,~,L] = unique(I(Id4(all(c(Id3),2),:)),'rows');
    P = histc(L, 1:size(D,1))/size(L,1);
else
    error('Decomposition by %dx%d is too big!', N, N);
end

function H = Hp(P)
% Entropy of a state of n micro-states
% P: micro-states probabilities vector (whatever in line or in column)
if abs(sum(P)-1) > 100000*eps
    error('Sum of probabilities not equal to 1: %0.4f', sum(P));
end
Ok = (P > eps) & (P < 1-eps);    % avoids NaN as p.log(p) -> 0 in 0+
H = sum(P(Ok).*log2(1./P(Ok)));  % normalization in bits
